/* A Bison parser, made by GNU Bison 3.8.2.  */

/* Bison interface for Yacc-like parsers in C

   Copyright (C) 1984, 1989-1990, 2000-2015, 2018-2021 Free Software Foundation,
   Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */

/* As a special exception, you may create a larger work that contains
   part or all of the Bison parser skeleton and distribute that work
   under terms of your choice, so long as that work isn't itself a
   parser generator using the skeleton or a modified version thereof
   as a parser skeleton.  Alternatively, if you modify or redistribute
   the parser skeleton itself, you may (at your option) remove this
   special exception, which will cause the skeleton and the resulting
   Bison output files to be licensed under the GNU General Public
   License without this special exception.

   This special exception was added by the Free Software Foundation in
   version 2.2 of Bison.  */

/* DO NOT RELY ON FEATURES THAT ARE NOT DOCUMENTED in the manual,
   especially those whose name start with YY_ or yy_.  They are
   private implementation details that can be changed or removed.  */

#ifndef YY_YY_SOAPCPP__YACC_H_INCLUDED
# define YY_YY_SOAPCPP__YACC_H_INCLUDED
/* Debug traces.  */
#ifndef YYDEBUG
# define YYDEBUG 0
#endif
#if YYDEBUG
extern int yydebug;
#endif

/* Token kinds.  */
#ifndef YYTOKENTYPE
# define YYTOKENTYPE
  enum yytokentype
  {
    YYEMPTY = -2,
    YYEOF = 0,                     /* "end of file"  */
    YYerror = 256,                 /* error  */
    YYUNDEF = 257,                 /* "invalid token"  */
    PRAGMA = 258,                  /* PRAGMA  */
    AUTO = 259,                    /* AUTO  */
    DOUBLE = 260,                  /* DOUBLE  */
    INT = 261,                     /* INT  */
    STRUCT = 262,                  /* STRUCT  */
    BREAK = 263,                   /* BREAK  */
    ELSE = 264,                    /* ELSE  */
    LONG = 265,                    /* LONG  */
    SWITCH = 266,                  /* SWITCH  */
    CASE = 267,                    /* CASE  */
    ENUM = 268,                    /* ENUM  */
    REGISTER = 269,                /* REGISTER  */
    TYPEDEF = 270,                 /* TYPEDEF  */
    CHAR = 271,                    /* CHAR  */
    EXTERN = 272,                  /* EXTERN  */
    RETURN = 273,                  /* RETURN  */
    UNION = 274,                   /* UNION  */
    CONST = 275,                   /* CONST  */
    FLOAT = 276,                   /* FLOAT  */
    SHORT = 277,                   /* SHORT  */
    UNSIGNED = 278,                /* UNSIGNED  */
    CONTINUE = 279,                /* CONTINUE  */
    FOR = 280,                     /* FOR  */
    SIGNED = 281,                  /* SIGNED  */
    VOID = 282,                    /* VOID  */
    DEFAULT = 283,                 /* DEFAULT  */
    GOTO = 284,                    /* GOTO  */
    SIZEOF = 285,                  /* SIZEOF  */
    VOLATILE = 286,                /* VOLATILE  */
    DO = 287,                      /* DO  */
    IF = 288,                      /* IF  */
    STATIC = 289,                  /* STATIC  */
    WHILE = 290,                   /* WHILE  */
    CLASS = 291,                   /* CLASS  */
    PRIVATE = 292,                 /* PRIVATE  */
    PROTECTED = 293,               /* PROTECTED  */
    PUBLIC = 294,                  /* PUBLIC  */
    VIRTUAL = 295,                 /* VIRTUAL  */
    INLINE = 296,                  /* INLINE  */
    OPERATOR = 297,                /* OPERATOR  */
    LLONG = 298,                   /* LLONG  */
    BOOL = 299,                    /* BOOL  */
    CFALSE = 300,                  /* CFALSE  */
    CTRUE = 301,                   /* CTRUE  */
    WCHAR = 302,                   /* WCHAR  */
    TIME = 303,                    /* TIME  */
    USING = 304,                   /* USING  */
    NAMESPACE = 305,               /* NAMESPACE  */
    ULLONG = 306,                  /* ULLONG  */
    MUSTUNDERSTAND = 307,          /* MUSTUNDERSTAND  */
    SIZE = 308,                    /* SIZE  */
    FRIEND = 309,                  /* FRIEND  */
    TEMPLATE = 310,                /* TEMPLATE  */
    EXPLICIT = 311,                /* EXPLICIT  */
    TYPENAME = 312,                /* TYPENAME  */
    RESTRICT = 313,                /* RESTRICT  */
    null = 314,                    /* null  */
    UCHAR = 315,                   /* UCHAR  */
    USHORT = 316,                  /* USHORT  */
    UINT = 317,                    /* UINT  */
    ULONG = 318,                   /* ULONG  */
    NONE = 319,                    /* NONE  */
    ID = 320,                      /* ID  */
    TAG = 321,                     /* TAG  */
    LAB = 322,                     /* LAB  */
    TYPE = 323,                    /* TYPE  */
    LNG = 324,                     /* LNG  */
    DBL = 325,                     /* DBL  */
    CHR = 326,                     /* CHR  */
    STR = 327,                     /* STR  */
    PA = 328,                      /* PA  */
    NA = 329,                      /* NA  */
    TA = 330,                      /* TA  */
    DA = 331,                      /* DA  */
    MA = 332,                      /* MA  */
    AA = 333,                      /* AA  */
    XA = 334,                      /* XA  */
    OA = 335,                      /* OA  */
    LA = 336,                      /* LA  */
    RA = 337,                      /* RA  */
    OR = 338,                      /* OR  */
    AN = 339,                      /* AN  */
    EQ = 340,                      /* EQ  */
    NE = 341,                      /* NE  */
    LE = 342,                      /* LE  */
    GE = 343,                      /* GE  */
    LS = 344,                      /* LS  */
    RS = 345,                      /* RS  */
    AR = 346,                      /* AR  */
    PP = 347,                      /* PP  */
    NN = 348                       /* NN  */
  };
  typedef enum yytokentype yytoken_kind_t;
#endif
/* Token kinds.  */
#define YYEMPTY -2
#define YYEOF 0
#define YYerror 256
#define YYUNDEF 257
#define PRAGMA 258
#define AUTO 259
#define DOUBLE 260
#define INT 261
#define STRUCT 262
#define BREAK 263
#define ELSE 264
#define LONG 265
#define SWITCH 266
#define CASE 267
#define ENUM 268
#define REGISTER 269
#define TYPEDEF 270
#define CHAR 271
#define EXTERN 272
#define RETURN 273
#define UNION 274
#define CONST 275
#define FLOAT 276
#define SHORT 277
#define UNSIGNED 278
#define CONTINUE 279
#define FOR 280
#define SIGNED 281
#define VOID 282
#define DEFAULT 283
#define GOTO 284
#define SIZEOF 285
#define VOLATILE 286
#define DO 287
#define IF 288
#define STATIC 289
#define WHILE 290
#define CLASS 291
#define PRIVATE 292
#define PROTECTED 293
#define PUBLIC 294
#define VIRTUAL 295
#define INLINE 296
#define OPERATOR 297
#define LLONG 298
#define BOOL 299
#define CFALSE 300
#define CTRUE 301
#define WCHAR 302
#define TIME 303
#define USING 304
#define NAMESPACE 305
#define ULLONG 306
#define MUSTUNDERSTAND 307
#define SIZE 308
#define FRIEND 309
#define TEMPLATE 310
#define EXPLICIT 311
#define TYPENAME 312
#define RESTRICT 313
#define null 314
#define UCHAR 315
#define USHORT 316
#define UINT 317
#define ULONG 318
#define NONE 319
#define ID 320
#define TAG 321
#define LAB 322
#define TYPE 323
#define LNG 324
#define DBL 325
#define CHR 326
#define STR 327
#define PA 328
#define NA 329
#define TA 330
#define DA 331
#define MA 332
#define AA 333
#define XA 334
#define OA 335
#define LA 336
#define RA 337
#define OR 338
#define AN 339
#define EQ 340
#define NE 341
#define LE 342
#define GE 343
#define LS 344
#define RS 345
#define AR 346
#define PP 347
#define NN 348

/* Value type.  */
#if ! defined YYSTYPE && ! defined YYSTYPE_IS_DECLARED
union YYSTYPE
{
#line 125 "soapcpp2_yacc.y"
	Symbol	*sym;
	LONG64	i;
	double	r;
	char	c;
	char	*s;
	Tnode	*typ;
	Storage	sto;
	Node	rec;
	Entry	*e;

#line 264 "soapcpp2_yacc.h"

};
typedef union YYSTYPE YYSTYPE;
# define YYSTYPE_IS_TRIVIAL 1
# define YYSTYPE_IS_DECLARED 1
#endif


extern YYSTYPE yylval;


int yyparse (void);


#endif /* !YY_YY_SOAPCPP__YACC_H_INCLUDED  */
